int read_field(FILE *fp, char *buff, signed *len)
{
  int c;
  unsigned n = 0;

  c = fgetc(fp);
  if(c != '"')
  {
    *len = -1;
    return c;
  }

  for(*len=0;c != EOF && n < CELLVAL_LEN-1;++*len)
  {
    c = fgetc(fp);
    if(c == '"')
    {
      c = fgetc(fp);
      if(c != '"')
        break;
    }
    buff[n++] = c;
  }
  buff[n] = 0;
  return c;
}

int add_cell(struct sheet *sheet, unsigned x, unsigned y, char *value);
void draw_table(struct sheet *sheet);
void set_cell_color(struct sheet *, unsigned col, unsigned row, char color);

int calculate_all_cells(struct sheet *sheet)
{
  unsigned row, col;
  struct cell *cell;
  int rc;

  show_calculating();

  for(row=1;row<=sheet->rows;row++)
    for(col=1;col<=sheet->lines[row-1].cell_count;col++)
    {
      if(
#ifdef DOS
kbhit() ||
#endif
sigint_received) return 0;
      cell = ptr_to_cell(sheet, col, row);
      if(cell)
      {
        set_cell_color(sheet, col, row, 0x70);
        if(rc = calculate_cell_result(sheet, cell))
        {
          if(rc == 1)
          {
            recalculate_refs(sheet, cell);
            clean_vars();
            draw_table(sheet);
            refresh();
          }
          else goto restorecolor;
        }
        else
          restorecolor:set_cell_color(sheet, col, row, 0x07);
      }
    }
}

int load_file(char *filename, struct sheet *sheet)
{
  FILE *fp = fopen(filename, "rb");
  unsigned row, col;
  char buff[CELLVAL_LEN];
  int c;
  signed clen;

  if(!fp)
    return -1;

  for(row=1;;row++)
  {
    for(col=1;;col++)
    {
      c = read_field(fp, buff, &clen);
      if(c == EOF) goto end;
      if(clen >= 0)
        if(add_cell(sheet, col, row, buff)) return -1;
      if(c == 0x0D)
      {
        c = fgetc(fp);
        break;
      }
    }
    if(c == EOF) break;
  }

  end:
  fclose(fp);

  strcpy(sheet->filename, filename);

  return 0;
}

int save_cell_to_file(char *value, FILE *fp)
{
  int c;
  fputc('"', fp);
  while(*value)
  {
    c = fputc(*value++, fp);
    if(c == '"') fputc(c, fp);
  }
  fputc('"', fp);
}

int save_file(char *filename, struct sheet *sheet, unsigned res)
{
  FILE *fp = fopen(filename, "wb");
  unsigned row, col;
  struct cell *cell;
  int c;
  char result[20];

  if(!fp)
    return -1;

  for(row=0;row<sheet->rows;row++)
  {
    for(col=0;col < sheet->lines[row].cell_count;col++)
    {
      if(col) fputc(',', fp);
      cell = ptr_to_cell(sheet, col+1, row+1);
      if(cell)
      {
        if(!res || cell->type == CELLTYPE_STR)
          save_cell_to_file(cell->value, fp);
        else
        {
          if(cell->restype == CELLTYPE_INT)
            if(cell->len == CELLLEN_SHORT)
              sprintf(result, "%i", cell->result.shortval);
            else if(cell->len == CELLLEN_LONG)
              sprintf(result, "%li", cell->result.longval);
            else
              sprintf(result, "%lli", cell->result.llongval);
          else
            sprintf(result, "%g", cell->result.doubleval);
          save_cell_to_file(result, fp);
        }
      }
      else
        save_cell_to_file("", fp);
    }
    fputc(0x0D, fp);
    fputc(0x0A, fp);
  }

  return fclose(fp);
}

int get_user_string(char *str, unsigned len, unsigned n, char *prompt);

int save_file_2(struct sheet *sheet, unsigned res)
{
  char filename[FILENAME_LEN+1] = { 0 };
  int rc;

  if(get_user_string(filename, sizeof(filename)-1, 0, "Enter file name"))
    return 0;
  rc = save_file(filename, sheet, res);
  if(!rc && !res)
    strcpy(sheet->filename, filename);
  return rc;
}
