#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef POSIX
#include <curses.h>
#include <unistd.h>
#ifndef UNIX
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#endif
#endif
#ifdef DOS
#include <conio.h>
#include <malloc.h>
#endif

#include <string.h>
#include <limits.h>
#include <errno.h>
#include <signal.h>

#include "structs.c"
#include "globals.c"

char statuslinebuff[60+CELLVAL_LEN];

#ifdef POSIX
#include "linux.c"
#endif

#ifdef DOS
#include "dos.c"
#endif

#include "signals.c"
#include "common.c"
#include "vars.c"
#include "macros.c"
#include "math.c"
#include "loadsave.c"

#define CELL_WIDTH 12

void show_status(char *str)
{
  status_line(str);
  refresh();
}

void show_status_getch(char *str)
{
  show_status(str);
  getch();
}

void num_to_letter(unsigned num, char *buff)
{
  unsigned o;

  memset(buff, ' ', CELL_WIDTH-1);
  buff[CELL_WIDTH] = 0;
  buff[CELL_WIDTH-1] = 'A';

  while(num--)
  {
    o = CELL_WIDTH-1;
    while(buff[o]++ == 'Z')
    {
      buff[o--] = 'A';
      if(buff[o] == ' ')
      {
        buff[o] = 'A';
        break;
      }
    }
  }
}


struct cell *ptr_to_cell(struct sheet *sheet, unsigned x, unsigned y);

char *cell_datatypes[] = 
{
  "string",
  "function",
  "integer",
  "float"
};

void set_cell_color(struct sheet *sheet, unsigned col, unsigned row, char color)
{
  int sr = row - sheet->scroll_row;
  int sc = col - sheet->scroll_col;

  if(sr > 0
    && sc > 0
    && sc <= scr_cols/CELL_WIDTH && sr <= scr_rows-2)
#ifndef DOS
  set_color(8+(sc-1)*CELL_WIDTH, sr+1, color, CELL_WIDTH);
  refresh();
#else
  if(color == 0x70 || color == 0x07)
    set_color_direct(8+(sc-1)*CELL_WIDTH, sr+1, color, CELL_WIDTH);
  else
  {
    set_color(8+(sc-1)*CELL_WIDTH, sr+1, color, CELL_WIDTH);
    refresh();
  }
#endif

}

void draw_table(struct sheet *sheet)
{
  unsigned n;
  unsigned r;
  char cell_text[CELLVAL_LEN];
  struct cell *cell;
  struct line *line;
//  signed sr, sc;

  if(current_for_cell) return;

  print_line(0, 1, sheet->filename, 0x20);

  for(n=0;n<scr_cols/CELL_WIDTH;n++)
  {
    num_to_letter(n+sheet->scroll_col, cell_text);
    print_line(8+n*CELL_WIDTH, 1, cell_text, 0x20);
  }
  for(n=0;n<scr_rows-3;n++)
  {
    sprintf(cell_text, "% 8u", n+sheet->scroll_row+1);
    print_line(0, n+2, cell_text, 0x20);
  }

  for(r=0;r+sheet->scroll_row < sheet->rows && r < scr_rows-3;r++)
  {
    line = &sheet->lines[r+sheet->scroll_row];
    for(n=0;n+sheet->scroll_col < line->cell_count && n < scr_cols/CELL_WIDTH;n++)
    {
      if(cell = line->cells[n+sheet->scroll_col])
      {
        if(cell->type == CELLTYPE_STR)
          sprintf(cell_text, "% 12s", cell->value);
        else
        {
          if(cell->restype == CELLTYPE_INT)
            if(cell->len == CELLLEN_SHORT)
              sprintf(cell_text, "% 12i", cell->result.shortval);
            else if(cell->len == CELLLEN_LONG)
              sprintf(cell_text, "% 12li", cell->result.longval);
            else
              sprintf(cell_text, "% 12lli", cell->result.llongval);
          else
            sprintf(cell_text, "% 12g", cell->result.doubleval);
        }
        print_line(8+n*CELL_WIDTH, 2+r, cell_text, cell->error ? 0x40 : 0x07);
      }
    }
  }

/*  if((sr = sheet->selected_row - sheet->scroll_row) > 0
     && (sc = sheet->selected_col - sheet->scroll_col) > 0
     && sc <= scr_cols/CELL_WIDTH && sr <= scr_rows-2)
  {
    set_cell_color(sc, sr-1, 0x3F);
  }*/

  num_to_letter(sheet->selected_col-1, cell_text);
  sprintf(statuslinebuff, "%s%u = ", cell_text, sheet->selected_row);

  if(cell = ptr_to_cell(sheet, sheet->selected_col, sheet->selected_row))
  {
    strcat(statuslinebuff, cell->value);
    sprintf(cell_text, " (%s)", cell_datatypes[cell->type]);
    strcat(statuslinebuff, cell_text);
  }
  else
    strcat(statuslinebuff, "(empty)");

}

void draw_screen(struct sheet *sheet)
{
  print_line(0, 0, "[F2 Save] [F3 Find next] [F4 Find first] [F5 Recalc] [F6 Save results] [F7 Jump]", 0x07);

  draw_table(sheet);

  show_status(statuslinebuff);

  set_cell_color(sheet, sheet->selected_col, sheet->selected_row, 0x3F);
}

struct sheet *new_sheet(void)
{
  struct sheet *retval = calloc(1, sizeof(struct sheet));
  if(!retval)
    return 0;

  retval->selected_col = retval->selected_row = 1;

  return retval;
}

int add_cell(struct sheet *sheet, unsigned x, unsigned y, char *value)
{
  void *ptr;
  struct cell *cell;

  if(y > sheet->rows)
  {
    if(!sheet->lines)
    {
      sheet->lines = calloc(1, y*sizeof(struct line));
      if(!sheet->lines) return -1;
    }
    else
    {
      ptr = realloc(sheet->lines, y*sizeof(struct line));
      if(!ptr) return -1;
      sheet->lines = ptr;
      memset(&sheet->lines[sheet->rows], 0, (y-sheet->rows)*sizeof(struct line));
    }
    sheet->rows = y;
  }

  if(x > sheet->lines[y-1].cell_count)
  {
    if(!sheet->lines[y-1].cells)
    {
      sheet->lines[y-1].cells = calloc(1, x*sizeof(struct cell *));
      if(!sheet->lines[y-1].cells) return -1;
    }
    else
    {
      ptr = realloc(sheet->lines[y-1].cells, x*sizeof(struct cell *));
      if(!ptr) return -1;
      sheet->lines[y-1].cells = ptr;
      memset(&sheet->lines[y-1].cells[sheet->lines[y-1].cell_count], 0, (x-sheet->lines[y-1].cell_count)*sizeof(struct cell*));
    }
    sheet->lines[y-1].cell_count = x;
  }

  if(!(sheet->lines[y-1].cells[x-1] = calloc(1, sizeof(struct cell))))
    return -1;

  cell = sheet->lines[y-1].cells[x-1];

//  strcpy(sheet->lines[y-1].cells[x-1]->value, value);
  strcpy(cell->value, value);
  cell->coords.row = y;
  cell->coords.col = x;
  

  return 0;
}

struct cell *ptr_to_cell(struct sheet *sheet, unsigned x, unsigned y)
{
  if(x && y && sheet->rows >= y && sheet->lines[y-1].cell_count >= x)
    return sheet->lines[y-1].cells[x-1];
  return 0;
}

void select_cell(struct sheet *sheet, unsigned x, unsigned y)
{
  sheet->selected_row = y;
  sheet->selected_col = x;
}

int get_user_string(char *str, unsigned len, unsigned n, char *prompt)
{
  int c;


  while(1)
  {
    sprintf(statuslinebuff, "%s: %s", prompt, str);
    show_status(statuslinebuff);

    c = getkeypress();
    if(c == TEXTEDIT_KEY_ENTER)
      break;
    if(c == TEXTEDIT_KEY_ESC)
      return -1;
    else if(c == TEXTEDIT_KEY_BACKSPACE)
    {
      if(n)
        str[--n] = 0;
      continue;
    }
    if(n+1 == len) continue;

    str[n++] = c;
    str[n] = 0;

  }
  return 0;
}

int edit_cell(struct sheet *sheet)
{
  unsigned n;
  struct cell *cell;
  char buf[CELLVAL_LEN];

  reset_lists();

  if(!(cell = ptr_to_cell(sheet, sheet->selected_col, sheet->selected_row)))
  {
    if(add_cell(sheet, sheet->selected_col, sheet->selected_row, ""))
      return -1;
    cell = ptr_to_cell(sheet, sheet->selected_col, sheet->selected_row);
  }
  else
  {
    free_deps(sheet, cell);
    memset(&cell->result, 0, sizeof(union cell_result));
  }

  strcpy(buf, cell->value);
  n = strlen(buf);

  get_user_string(buf, CELLVAL_LEN, n, "Enter new value");

  strcpy(ptr_to_cell(sheet, sheet->selected_col, sheet->selected_row)->value, buf);

  show_calculating();
  calculate_cell_result(sheet, cell);
  recalculate_refs(sheet, cell);
  clean_vars();

  return 0;
}

void jump_to_row(struct sheet *sheet)
{
  char buf[7];
  struct coords coords;
  char *ptr;

  buf[0] = 0;

  if(get_user_string(buf, sizeof(buf), 0, "Jump to cell"))
    return;
  str_to_coords(buf, &coords, &ptr);
  sheet->selected_col = coords.col;
  sheet->selected_row = coords.row;
}

char searchbuf[CELLVAL_LEN] = { 0 };

void search_text(struct sheet *sheet, unsigned start_col, unsigned start_row)
{
  unsigned row, col;

  for(row=start_row,col=start_col;row<sheet->rows;row++,col=0)
    for(;col<sheet->lines[row].cell_count;col++)
    {
      if(sheet->lines[row].cells[col] && strstr(sheet->lines[row].cells[col]->value, searchbuf))
      {
        sheet->selected_row = row+1;
        sheet->selected_col = col+1;
        return;
      }
    }
  show_status_getch("Not found");
}

void search_next(struct sheet *sheet)
{
  search_text(sheet, sheet->selected_col, sheet->selected_row-1);
}

void search_first(struct sheet *sheet)
{
  if(get_user_string(searchbuf, sizeof(searchbuf), strlen(searchbuf), "Enter search word"))
    return;

  search_text(sheet, 0, 0);
}

void delete_cell(struct sheet *sheet)
{
  unsigned n;
  struct line *line = &sheet->lines[sheet->selected_row-1];
  struct cell *cell = ptr_to_cell(sheet, sheet->selected_col, sheet->selected_row);

  if(!cell) return;

  memset(&cell->result, 0, sizeof(union cell_result));
  recalculate_refs(sheet, cell);

  free_deps(sheet, cell);
  free_refs(sheet, cell);

  free(cell);
  line->cells[sheet->selected_col-1] = 0;
  recalculate_range(sheet, sheet->selected_row);

//  for(n=sheet->selected_col;n<=line->cell_count;n++)
//    if(ptr_to_cell(sheet, n, sheet->selected_row))
//      return;
//  line->cell_count = sheet->selected_col-1;
  for(n=line->cell_count;n;n--)
    if(ptr_to_cell(sheet, n, sheet->selected_row))
      break;
  line->cell_count = n;
}

#ifndef UNIX
int scroll_lock_enabled(void)
{
#ifdef POSIX
  int rc;
  char state;
  int fd = 0;
  if(ioctl(fd, KDGKBLED, &state) == -1)
    rc = 0;
  else if(state & LED_SCR)
    rc = 1;
  else
    rc = 0;
  return rc;
#endif
#ifdef DOS
  return !!((*(unsigned char __far*)0x00400097) & 0x01);
#endif
}
#endif

int user_loop(struct sheet *sheet)
{
  int c;
  unsigned rc, br;
  signed *row, *col;
  int sl;

  while(1)
  {
    sigint_received = 0;
    parser_callstack = 0;
    clear_screen();
    draw_screen(sheet);
    c = getkeypress();

#ifndef UNIX
    if(!(sl = scroll_lock_enabled()))
    {
#endif
      row = &sheet->selected_row;
      col = &sheet->selected_col;
#ifndef UNIX
    }
    else
    {
      row = &sheet->scroll_row;
      col = &sheet->scroll_col;
    }
#endif

    switch(c)
    {
      case TEXTEDIT_KEY_ESC: return 0;
      case TEXTEDIT_KEY_DOWN: ++*row; break;
      case TEXTEDIT_KEY_UP: if(*row) --*row; break;
      case TEXTEDIT_KEY_LEFT: if(*col) --*col; break;
      case TEXTEDIT_KEY_RIGHT: ++*col; break;
      case TEXTEDIT_KEY_PGUP: *row -= scr_rows; if(*row < 0) *row = 0; break;
      case TEXTEDIT_KEY_HOME: *col = 0; break;
      case TEXTEDIT_KEY_END: if(*row && *row <= sheet->rows) *col = sheet->lines[*row-1].cell_count;
                             else *col += scr_cols/CELL_WIDTH-2; break;
      case TEXTEDIT_KEY_PGDN: *row += scr_rows; break;
      case TEXTEDIT_KEY_ENTER: edit_cell(sheet); break;
      case TEXTEDIT_KEY_DELETE: delete_cell(sheet); break;
      case TEXTEDIT_KEY_F2: save_file_2(sheet,0); break;
      case TEXTEDIT_KEY_F4: search_first(sheet); break;
      case TEXTEDIT_KEY_F3: search_next(sheet); break;
      case TEXTEDIT_KEY_F6: save_file_2(sheet, 1); break;
      case TEXTEDIT_KEY_F5: calculate_all_cells(sheet); break;
      case TEXTEDIT_KEY_F7: jump_to_row(sheet); break;
    }

    if(!sl)
    {
      if(!*row) ++*row;
      if(!*col) ++*col;

      rc = sheet->scroll_col + scr_cols/CELL_WIDTH;
      br = sheet->scroll_row + scr_rows - 3;

      if(sheet->selected_row >= br)
        sheet->scroll_row = sheet->selected_row - scr_rows + 3;
      else if(sheet->selected_row <= sheet->scroll_row)
        sheet->scroll_row = sheet->selected_row - 1;

      if(sheet->selected_col > rc)
        sheet->scroll_col = sheet->selected_col - scr_cols/CELL_WIDTH + 1;
      else if(sheet->selected_col <= sheet->scroll_col)
        sheet->scroll_col = sheet->selected_col - 1;
    }
  }
}

int main(int argc, char *argv[])
{
  struct sheet *sheet;

  signal(SIGINT, sigint_handler);

  temp_cells.count = 0;

  if(init_macros())
  {
    puts("Could not initialize macros.");
    exit(1);
  }

  sheet = new_sheet();

  if(argc == 2)
    if(load_file(argv[1], sheet))
    {
      printf("Error while loading file %s: %s\n", argv[1], strerror(errno));
      exit(1);
    }

  init_screen();

  return user_loop(sheet);
}
