/*
  TODO:
*/

struct cell *ptr_to_cell(struct sheet *sheet, unsigned x, unsigned y);

unsigned letter_to_num(char *ltrs, char **ptr)
{
  unsigned n,p,i,t;
  unsigned res = 0;

  for(n=0;ltrs[n] >= 'A' && ltrs[n] <= 'Z';n++);

  *ptr = ltrs + n;

  for(i=0;n--;i++)
  {
    t = ltrs[i] - 'A' + 1;

    for(p=n;p--;)
      t *= 26;

    res += t;
  }

  return res;
}

int add_cell_to_list(struct celllist *list, struct cell *cell)
{
  void *ptr;

  if(!list->count)
  {
    list->cells = malloc(sizeof(struct cell *));
    if(!list->cells)
      return -1;
  }
  else
  {
    ptr = realloc(list->cells, (1+list->count)*sizeof(struct cell *));
    if(!ptr)
      return -1;
    list->cells = ptr;
  }
  list->cells[list->count++] = cell;
  return 0;
}

int add_column_to_list(struct celllist *list, struct sheet *sheet, unsigned col)
{
  unsigned n;
  struct cell *cell;
  int rc;
  for(n=1;n<=sheet->rows;n++)
  {
    if(cell = ptr_to_cell(sheet, col, n))
      rc += add_cell_to_list(list, cell);
  }
  return rc;
}

int add_row_to_list(struct celllist *list, struct sheet *sheet, unsigned row)
{
  unsigned n;
  struct cell *cell;
  int rc;

  for(n=1;n<=sheet->lines[row-1].cell_count;n++)
  {
    if(cell = ptr_to_cell(sheet, n, row))
      rc += add_cell_to_list(list, cell);
  }
  return rc;
}

int find_range_reference(struct line *line, struct coords *coords)
{
  unsigned n;

  for(n=line->range_ref_count;n--;)
    if(!memcmp(&line->range_refs[n], coords, sizeof(struct coords)))
      break;

  return n;
}

int add_range_reference(struct sheet *sheet, unsigned row, struct coords *coords)
{
  struct line *line = &sheet->lines[row-1];
  void *ptr;

  if(!coords->row) return 0;

  if(find_range_reference(line, coords) != -1) return 0;

  if(!line->range_ref_count)
  {
    line->range_refs = malloc(sizeof(struct coords));
    if(!line->range_refs) return -1;
  }
  else
  {
    ptr = realloc(line->range_refs, (1+line->range_ref_count)*sizeof(struct coords));
    if(!ptr)
      return -1;
    line->range_refs = ptr;
  }
  line->range_refs[line->range_ref_count++] = *coords;

  return 0;
}

void delete_range_reference(struct sheet *sheet, unsigned row, struct coords *coords)
{
  unsigned n;
  struct line *line = &sheet->lines[row];

  for(n=line->range_ref_count;n--;)
    if(!memcmp(&line->range_refs[n], coords, sizeof(struct coords)))
      break;
  if(n == -1) return;

  memmove(&line->range_refs[n], &line->range_refs[n+1], (--line->range_ref_count - n)*sizeof(struct coords));
  if(!line->range_ref_count)
    free(line->range_refs);
  else
    line->range_refs = realloc(line->range_refs, line->range_ref_count*sizeof(struct coords));
}

int add_range_to_list(struct celllist *list, struct sheet *sheet,
      unsigned col1, unsigned row1, unsigned col2, unsigned row2, struct cell *rcell)
{
  unsigned r, c;
  struct cell *cell;
  int rc = 0;
  struct coords *ref = &rcell->coords;

  if(!row1 && col1)
  {
    for(c=col1;c<=col2;c++)
      add_column_to_list(list, sheet, c);
    rcell->col_dep = 1;
  }
  else if(!col1)
  {
    for(r=row1;r<=row2;r++)
    {
      add_row_to_list(list, sheet, r);
      if(r <= sheet->rows)
        rc += add_range_reference(sheet, r, ref);
    }
  }
  else
  {
    for(r=row1;r<=row2;r++)
    {
      for(c=col1;c<=col2;c++)
        if(cell = ptr_to_cell(sheet, c, r))
          add_cell_to_list(list, cell);
      if(r <= sheet->rows)
        rc += add_range_reference(sheet, r, ref);
    }
  }
  return rc;
}


int add_ref(struct cell *cell, struct coords ref);
int add_dep(struct cell *cell, struct coords dep);
void free_temp_cells(void);

int save_temporary_deps(struct sheet *sheet, struct cell *cell)
{
  unsigned n = temp_cells.count;
  unsigned d;
  struct cell *tcell;
  struct cell *rcell;

  while(n--)
  {
    tcell = temp_cells.cells[n];
    for(d=tcell->dep_count;d--;)
    {
      rcell = ptr_to_cell(sheet, tcell->deps[d].col, tcell->deps[d].row);
      add_dep(cell, tcell->deps[d]);
      add_ref(rcell, cell->coords);
    }
    cell->col_dep |= tcell->col_dep;
    cell->error |= tcell->error;
  }

  free_temp_cells();
}

struct cell *new_temp_cell(void)
{
  void *ptr;
  struct cell *cell;

  cell = calloc(1, sizeof(struct temporary_cell));
  if(!cell) return 0;

  if(!temp_cells.count)
  {
    temp_cells.cells = malloc(sizeof(struct cell*));
    if(!temp_cells.cells)
    {
      outofmemory:
      free(cell);
      return 0;
    }
  }
  else
  {
    ptr = realloc(temp_cells.cells, (1+temp_cells.count)*sizeof(struct cell*));
    if(!ptr)
      goto outofmemory;
    temp_cells.cells = ptr;
  }
  temp_cells.cells[temp_cells.count++] = cell;
  cell->loopvar = program_status.loop_body_eval;

  return cell;
}

void free_temp_cells(void)
{
  unsigned n = temp_cells.count;

  while(n--)
  {
    free(temp_cells.cells[n]->refs);
    free(temp_cells.cells[n]->deps);
    free(temp_cells.cells[n]);
  }
  free(temp_cells.cells);
  temp_cells.cells = 0;
  temp_cells.count = 0;
}

int parse_function(struct sheet *sheet, struct cell *cell);

void free_deps(struct sheet *sheet, struct cell *cell);

int is_delimiter(char chr)
{
  if(chr == ',' || chr == ';')
    return 1;
  else return 0;
}

char *mempbrk(char *str, char *ndls, unsigned count)
{
  char c;
  char *ptr;
  while(c = *ndls++)
  {
    ptr = memchr(str, c, count);
    if(ptr)
      break;
  }
  return ptr;
}

int calculate_cell_result(struct sheet *sheet, struct cell *cell);

void str_to_coords(char *str, struct coords *coords, char **endptr)
{
  coords->col = letter_to_num(str, &str);
  coords->row = strtoul(str, endptr, 10);
}

struct cell *function_from_args(struct sheet *sheet, char *args, char **endp, char *pptr);

#define OPER_EQUALS 0x1
#define OPER_GREATER 0x2
#define OPER_LESS 0x4
#define OPER_NOTEQUAL 0x8
#define OPER_AND 0x10
#define OPER_OR 0x20
#define OPER_XOR 0x40
//#define CMP_FLT 0x80

int compare(struct logical_result *cval, struct cmpstrct *cmpstrct)
{
  unsigned i = 2;
  unsigned char oper = 0;
  long long res;
  float fres;

//  if(cval[0].flt || cval[1].flt)
//    oper |= CMP_FLT;
  oper = cval[0].flt | cval[1].flt;


  if(oper/* & CMP_FLT*/)
  {
    for(i=2;i--;)
    {
      if(cval[i].imm)
        if(cval[i].flt)
          cmpstrct[i].cmpflt = cval[i].value.floatval;
        else
          cmpstrct[i].cmpflt = cval[i].value.longval;
      else
        if(cval[i].flt)
          cmpstrct[i].cmpflt = cval[i].value.cellptr->result.doubleval;
        else
          if(cval[i].value.cellptr->len == CELLLEN_SHORT)
            cmpstrct[i].cmpflt = cval[i].value.cellptr->result.shortval;
          else if(cval[i].value.cellptr->len == CELLLEN_LONG)
            cmpstrct[i].cmpflt = cval[i].value.cellptr->result.longval;
          else if(cval[i].value.cellptr->len == CELLLEN_LLONG)
            cmpstrct[i].cmpflt = cval[i].value.cellptr->result.llongval;
    }
    fres = cmpstrct[1].cmpflt - cmpstrct[0].cmpflt;
    res = fres < 0 ? floor(fres) : ceil(fres);
  }
  else
  {
    for(i=2;i--;)
    {
      if(cval[i].imm)
        cmpstrct[i].cmpint = cval[i].value.longval;
      else
        if(cval[i].value.cellptr->len == CELLLEN_SHORT)
          cmpstrct[i].cmpint = cval[i].value.cellptr->result.shortval;
        else if(cval[i].value.cellptr->len == CELLLEN_LONG)
          cmpstrct[i].cmpint = cval[i].value.cellptr->result.longval;
        else if(cval[i].value.cellptr->len == CELLLEN_LLONG)
          cmpstrct[i].cmpint = cval[i].value.cellptr->result.llongval;
    }
    res = cmpstrct[1].cmpint - cmpstrct[0].cmpint;
  }
  if(!res) return 0;
  else if(res < 0) return -1;
  else return 1;
}

struct cell *parse_logical_expression(struct sheet *sheet, char *args, char **endptr)
{
  struct cell *retval = new_temp_cell();
  struct coords coords;
  struct cell *cell;
  struct logical_result cval[2];
  unsigned i = 2;
  unsigned char oper = 0;
  int res;
//  float fres;
  struct cmpstrct cmpstrct[2];
  char *ptr,*ptr2;

#define OPER_EQUALS 0x1
#define OPER_GREATER 0x2
#define OPER_LESS 0x4
#define OPER_NOTEQUAL 0x8
#define OPER_AND 0x10
#define OPER_OR 0x20
#define OPER_XOR 0x40
#define CMP_FLT 0x80

  if(!retval) return 0;

  ptr = args;

  while(i--)
  {
    ptr2 = strpbrk(args, "<=>;,");
    if(ptr2 && (ptr2 = memchr(args, '(', ptr2-args)))
    {
      cell = function_from_args(sheet, args, &args, ptr2);
      if(cell)
        goto ptrtocell;
      else
      {
        error:
        retval->error = 1;
        return retval;
      }
    }
    else if(*args >= 'A' && *args <= 'Z')
    {
      str_to_coords(args, &coords, &args);
      cell = ptr_to_cell(sheet, coords.col, coords.row);
      if(!cell)
      {
        cval[i].value.longval = 0;
        cval[i].flt = 0;
        goto setimm;
      }
      add_dep(retval, coords);
      ptrtocell:
      cval[i].value.cellptr = cell;
      cval[i].imm = 0;
      cval[i].flt = (cell->restype == CELLTYPE_FLT);
    }
    else if(*args == '$')
    {
      cell = get_var(*++args);
      args++;
      goto ptrtocell;
    }
    else
    {
      cval[i].value.longval = strtol(args, endptr, 10);
      if(**endptr != '.')
      {
        cval[i].flt = 0;
        args = *endptr;
      }
      else
      {
        cval[i].flt = 1;
        cval[i].value.floatval = strtod(args, &args);
      }
      setimm:cval[i].imm = 1;
    }
    if(i)
    {
      switch(*args++)
      {
        case '!': oper |= OPER_NOTEQUAL;
                  goto chkeq;
        case '>': oper |= OPER_GREATER;
                  chkeq:if(*args == '=') goto add_equals;
                  break;
        case '<': oper |= OPER_LESS;
                  if(*args != '=') break;
                  add_equals: args++;
        case '=': equals:oper |= OPER_EQUALS; break;
      }
    }
  }

  memcpy(retval->value, "=SUM(", 5);
  memcpy(&retval->value[5], ptr, args-ptr);
  retval->value[5+args-ptr] = ')';

/*  if(cval[0].flt || cval[1].flt)
      oper |= CMP_FLT;

  if(oper & CMP_FLT)
  {
    for(i=2;i--;)
    {
      if(cval[i].imm)
        if(cval[i].flt)
          cmpstrct[i].cmpflt = cval[i].value.floatval;
        else
          cmpstrct[i].cmpflt = cval[i].value.longval;
      else
        if(cval[i].flt)
          cmpstrct[i].cmpflt = cval[i].value.cellptr->result.doubleval;
        else
          if(cval[i].value.cellptr->len == CELLLEN_SHORT)
            cmpstrct[i].cmpflt = cval[i].value.cellptr->result.shortval;
          else if(cval[i].value.cellptr->len == CELLLEN_LONG)
            cmpstrct[i].cmpflt = cval[i].value.cellptr->result.longval;
          else if(cval[i].value.cellptr->len == CELLLEN_LLONG)
            cmpstrct[i].cmpflt = cval[i].value.cellptr->result.llongval;
    }
    fres = cmpstrct[1].cmpflt - cmpstrct[0].cmpflt;
    res = fres < 0 ? floor(fres) : ceil(fres);
  }
  else
  {
    for(i=2;i--;)
    {
      if(cval[i].imm)
        cmpstrct[i].cmpint = cval[i].value.longval;
      else
        if(cval[i].value.cellptr->len == CELLLEN_SHORT)
          cmpstrct[i].cmpint = cval[i].value.cellptr->result.shortval;
        else if(cval[i].value.cellptr->len == CELLLEN_LONG)
          cmpstrct[i].cmpint = cval[i].value.cellptr->result.longval;
        else if(cval[i].value.cellptr->len == CELLLEN_LLONG)
          cmpstrct[i].cmpint = cval[i].value.cellptr->result.llongval;
    }
    res = cmpstrct[1].cmpint - cmpstrct[0].cmpint;
  }*/
  res = compare(cval, cmpstrct);

  switch(oper & 0x7F)
  {
    case OPER_EQUALS:
      retval->result.shortval = !res;
    break;
    case OPER_EQUALS|OPER_GREATER:
      retval->result.shortval = (res >= 0);
    break;
    case OPER_GREATER:
      retval->result.shortval = (res > 0);
      break;
    case OPER_EQUALS|OPER_LESS:
      retval->result.shortval = (res <= 0);
    break;
    case OPER_LESS:
      retval->result.shortval = (res < 0);
    break;
    case OPER_NOTEQUAL|OPER_EQUALS:
      retval->result.shortval = !!res;
    break;
  }

  retval->restype = CELLTYPE_INT;
  retval->len = CELLLEN_SHORT;
  retval->type = CELLTYPE_FNC;

  *endptr = args;

  return retval;
}

char *skip_parenthesis(char *str)
{
  unsigned p = 0;
  char *ptr = str;
  char *ptr2;

  do
  {
    ptr2 = strpbrk(ptr, "()");
    if(ptr2)
    {
      if(*ptr2 == '(')
        p++;
      else if(p)
        p--;
      else break;
      ptr = ptr2 + 1;
    }
  }
  while(p);

  return ptr;
}

struct cell *parse_list(struct sheet *, char *arg, unsigned, char **);

struct cell *function_from_args(struct sheet *sheet, char *args, char **endp, char *pptr)
{
  struct cell *cell;
  char *ptr, *ptr2;
  unsigned n;

  if(!memcmp(args, "LIST(", 5))
    if(cell = parse_list(sheet, &args[5], 0, endp))
      return cell;
    else
      emptycell:return new_temp_cell();
  else if(!memcmp(args, "SLIST(", 6))
    if(cell = parse_list(sheet, &args[6], 1, endp))
      return cell;
    else goto emptycell;

  cell = new_temp_cell();
  if(!cell) return 0;
  cell->value[0] = '=';
  ptr2 = strchr(args, ')');
  if(!ptr2)
  {
    syntaxerror:
    *endp = 0;
    return 0;
  }
  ptr = skip_parenthesis(pptr);
  if(!ptr) goto syntaxerror;
  n = ptr-args;
  memcpy(&cell->value[1], args, n);
  if(parse_function(sheet, cell))
    cell->type = CELLTYPE_FNC;
  else goto syntaxerror;
  args += n;
  *endp = args;

  return cell;
}

struct celllist cells_from_args(struct sheet *sheet, char *args, struct cell *rcell, unsigned argcnt, char **endptr)
{
  struct celllist retval;
  unsigned long arg;
  struct coords coords[2];
  struct cell *cell;
  unsigned char r = 0;
  unsigned char a = 0;
  unsigned n;
  char *ptr,*ptr2,*ptr3;
  char chr;
  char p;

  retval.cells = 0;
  retval.count = 0;
  while(argcnt--)
  {
    coords[r].col = coords[r].row = 0;
    ptr = strpbrk(args, ",;");
    if(!ptr)
      ptr = args+strlen(args);
    n = ptr-args;

    ptr = memchr(args, '(', n);
    if(ptr)
    {
      ptr3 = skip_parenthesis(ptr);
      if(memchr("<=>", *ptr3, 3))
        goto compare;
      n = ptr-args;
    }
    ptr2 = mempbrk(args, "<=>", n);
    if(ptr && !ptr2)
    {
      cell = function_from_args(sheet, args, &args, ptr);
      if(!args)
        goto syntaxerror;
      if(!cell)
        goto outofmemory;
      add_cell_to_list(&retval, cell);
      goto range;
    }
    else if(ptr2)
    {
      compare:
      cell = parse_logical_expression(sheet, args, &args);
      if(cell)
        add_cell_to_list(&retval, cell);
      else goto outofmemory;
      goto range;
    }


    if(*args >= 'A' && *args <= 'Z')
    {
      str_to_coords(args, &coords[r], &args);
      if(*args == ':') goto range;
    }
    else if(*args >= '0' && *args <= '9' || *args == '-')
    {
      arg = strtoul(args, &ptr, 10);
      if(is_delimiter(*(args-1)) || is_delimiter(*(ptr))
         || *ptr == '.' || !a && *(ptr) == ')')
      {
        // immediate number
        cell = new_temp_cell();
        if(!cell) goto outofmemory;
        if(*ptr != '.')
          memcpy(cell->value, args, ptr-args);
        else
        {
          ptr = strpbrk(args, ",;)");
          if(!ptr)
          {
            syntaxerror:
            show_status_getch("Syntax error");
            goto error;
          }
          memcpy(cell->value, args, ptr-args);
        }
        calculate_cell_result(sheet, cell);
        add_cell_to_list(&retval, cell);
        args = ptr;
        goto range;
      }
      else if(*args >= '1')
      {
        args = ptr;
        coords[r].row = arg;
        if(!r)
          goto range;
      }
    }
    else if(*args == '$') // variable
    {
      if(*++args >= 'A' && *args <= 'Z')
        cell = get_var(*args);
      else goto syntaxerror;
      if(!cell)
        goto syntaxerror;
      add_cell_to_list(&retval, cell);
      args=strpbrk(args, ",;)");
      if(!args) return retval;
      goto range;
    }

    if(r)
    {
      r = 0;
      if(add_range_to_list(&retval, sheet, coords[0].col, coords[0].row, coords[1].col, coords[1].row, rcell))
        goto outofmemory;
    }

    else if(cell = ptr_to_cell(sheet, coords[0].col, coords[0].row))
    {
      if(add_cell_to_list(&retval, cell))
      {
        outofmemory:
        show_status_getch("Out of memory");
        error:
        sigint_received = 1;
        free(retval.cells);
        retval.cells = 0;
        retval.count = -1;
        free_deps(sheet, rcell);
        return retval;
      }
    }
    else
    {
      invalidcell:show_status_getch("Invalid cell");
      goto error;
    }

    range:switch(*args)
    {
      case ')': case 0: goto end;
      case ',': case ';': args++; break;
      case ':':
        if(r)
        {
          show_status_getch("Invalid range");
          goto error;
        }
        args++;
        r++;
        break;
    }
    a++;
  }

  end:

  if(endptr)
    *endptr = args;

  return retval;
}

int find_coord_in_list(struct coords *list, struct coords ndl, unsigned count)
{
  while(count--)
    if(!memcmp(&list[count], &ndl, sizeof(struct coords)))
      break;
  return count;
}

int add_ref(struct cell *cell, struct coords ref)
{
  void *ptr;

  if(!cell->ref_count)
  {
    cell->refs = malloc(sizeof(struct coords));
    if(!cell->refs) return -1;
  }
  else
  {
    if(find_coord_in_list(cell->refs, ref, cell->ref_count) != -1)
      return 0;
    ptr = realloc(cell->refs, (cell->ref_count+1)*sizeof(struct coords));
    if(!ptr) return -1;
    cell->refs = ptr;
  }
  cell->refs[cell->ref_count++] = ref;

  return 0;
}

int add_dep(struct cell *cell, struct coords dep)
{
  void *ptr;

  if(!cell->dep_count)
  {
    cell->deps = malloc(sizeof(struct coords));
    if(!cell->deps) return -1;
  }
  else
  {
    if(find_coord_in_list(cell->deps, dep, cell->dep_count) != -1)
      return 0;
    ptr = realloc(cell->deps, (cell->dep_count+1)*sizeof(struct coords));
    if(!ptr) return -1;
    cell->deps = ptr;
  }
  cell->deps[cell->dep_count++] = dep;

  return 0;
}

int add_refs(struct celllist *cells, struct cell *cell)
{
  unsigned n;

//  if(!cell->coords.row) return 0;
  if(current_for_cell) return 0;

  for(n=cells->count;n--;)
  {
    if(cell->coords.row)
      if(add_ref(cells->cells[n], cell->coords))
      {
        error:
        free(cells->cells);
        return -1;
      }
    if(cells->cells[n]->coords.row)
      if(add_dep(cell, cells->cells[n]->coords)) goto error;
  }

  return 0;
}

void delete_dep(struct cell *cell, struct coords dep);

void delete_ref(struct cell *cell, struct coords *ref)
{
  unsigned n;

  for(n=cell->ref_count;n--;)
    if(!memcmp(&cell->refs[n], ref, sizeof(struct coords))) break;
  if(n == -1) return;

  memmove(&cell->refs[n], &cell->refs[n+1], (--cell->ref_count - n)*sizeof(struct coords));
  if(cell->ref_count)
    cell->refs = realloc(cell->refs, cell->ref_count*sizeof(struct coords));
  else
  {
    free(cell->refs);
    cell->refs = 0;
  }
}

// delete all dependencies from this cell
void free_deps(struct sheet *sheet, struct cell *cell)
{
  unsigned n;
  struct cell *dep;

  for(n=sheet->rows;n--;)
    delete_range_reference(sheet, n, &cell->coords);

  for(n=cell->dep_count;n--;)
  {
    dep = ptr_to_cell(sheet, cell->deps[n].col, cell->deps[n].row);
    if(dep)
      delete_ref(dep, &cell->coords);
  }

  cell->dep_count = 0;
  free(cell->deps);
  cell->deps = 0;
  cell->col_dep = 0;

}

void delete_dep(struct cell *cell, struct coords dep)
{
  unsigned n;

  for(n=cell->dep_count;n--;)
    if(!memcmp(&cell->deps[n], &dep, sizeof(struct coords))) break;
  if(n == -1) return;

  memmove(&cell->deps[n], &cell->deps[n+1], (--cell->dep_count - n)*sizeof(struct coords));
  if(cell->dep_count)
    cell->deps = realloc(cell->deps, cell->dep_count*sizeof(struct coords));
  else
  {
    free(cell->deps);
    cell->deps = 0;
  }
}

// delete all references to this cell
void free_refs(struct sheet *sheet, struct cell *cell)
{
  unsigned n;
  struct cell *ref;

  for(n=cell->ref_count;n--;)
  {
    ref = ptr_to_cell(sheet, cell->refs[n].col, cell->refs[n].row);
    if(ref)
      delete_dep(ref, cell->coords);
  }

  cell->ref_count = 0;
  free(cell->refs);
  cell->refs = 0;
}

int is_true(struct cell *cell)
{
  if((cell->restype == CELLTYPE_FLT && cell->result.doubleval)
     || (cell->restype == CELLTYPE_INT &&
       ((cell->len == CELLLEN_SHORT && cell->result.shortval)
       || (cell->len == CELLLEN_LONG && cell->result.longval)
       || (cell->len == CELLLEN_LLONG && cell->result.llongval))))
    return 1;
  return 0;
}

int parse_if(struct sheet *sheet, struct cell *cell)
{
  char *arg = &cell->value[4];
  struct celllist cells = cells_from_args(sheet, arg, cell, 1, &arg);
  unsigned n;
  struct cell *srccell;
  int rc;
  char *ptr;

  if(cells.count == -1) return 0;

  if(cells.count < 1 || !arg || !*arg) rc = 0;
  else
  {
    n = is_true(cells.cells[0]);
    free(cells.cells);
    cells.cells = 0;
    if(n)
    {
      cellfromarg:
      cells = cells_from_args(sheet, arg, cell, 1, 0);
      if(cells.count == -1) return 0;
      if(cells.count < 1) rc = 0;
      else
      {
        cell->result = cells.cells[0]->result;
        cell->restype = cells.cells[0]->restype;
        cell->len = cells.cells[0]->len;
        setrc:rc = 1;
      }
    }
    else
    {
      while(1)
      {
        ptr = strpbrk(arg, ";,");
        if(!ptr)
        {
          memset(&cell->result, 0, sizeof(union cell_result));
          cell->restype = CELLTYPE_INT;
          cell->len = CELLLEN_SHORT;
          goto setrc;
        }
        else
        {
          if(arg = memchr(arg, '(', ptr-arg))
            arg = skip_parenthesis(arg);
          else
          {
            arg = ptr+1;
            goto cellfromarg;
          }
        }
      }
    }
  }

  free(cells.cells);

  return rc;
}

void delete_temporary_cell(struct sheet *sheet, unsigned cell_id)
{
  struct cell *cell = temp_cells.cells[cell_id];
  free_deps(sheet, cell);
  free_refs(sheet, cell);

  free(cell);

  memmove(&temp_cells.cells[cell_id], &temp_cells.cells[cell_id+1], (--temp_cells.count-cell_id)*sizeof(struct cell*));
  temp_cells.cells = realloc(temp_cells.cells, temp_cells.count * sizeof(struct cell *));
}

int cell_in_list(struct cell *cell, struct celllist *list)
{
  unsigned n;
  for(n=list->count;n--;)
    if(list->cells[n] == cell)
      return 1;
  return 0;
}

// delete all temporary cells that are FOR variables, but leave cells that are in the list
void free_for_cells(struct sheet *sheet, struct celllist *list1, struct celllist *list2)
{
  unsigned n;
  struct cell *cell;

  for(n=temp_cells.count;n--;)
  {
    cell = temp_cells.cells[n];
    if(cell->loopvar && !cell_in_list(cell, list1) && !cell_in_list(cell, list2))
      delete_temporary_cell(sheet, n);
  }
}

int recalculate_cell(struct sheet *sheet, struct cell *cell);

void reset_lists(void);

int parse_for(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist initcells;
  struct celllist condcells;
  struct celllist bodycells;
  unsigned n;
  char *ptr;
  char *ptr2;
  char *condition;
  char loop_status = 0;
  int rc;
#define FOR_BODYCELLS_KNOWN 0x01

  ptr = strchr(arg, ';');
  if(!ptr) return 0;

  reset_lists();

  if(memchr(arg, '(', ptr-arg))
  {
    ptr2 = arg;
    do
    {
      ptr2 = skip_parenthesis(ptr2);
      if(!ptr2) return 0;
    }
    while(*ptr2 == ',');
    ptr = ptr2;
  }

  initcells.cells = 0;
  if(ptr != arg)
  {
    *ptr = 0;
    initcells = cells_from_args(sheet, arg, cell, -1, 0);
    *ptr = ';';
    if(initcells.count == -1)
      return 0;
  }

  condition = ptr+1;


  arg = condition;

  ptr = strchr(arg, ';');
  if(!ptr)
  {
    error1:
    free(initcells.cells);
    return 0;
  }

  if(memchr(arg, '(', ptr-arg))
  {
    ptr2 = arg;
    do
    {
      ptr2 = skip_parenthesis(ptr2);
      if(!ptr2)
        goto error1;
    }
    while(*ptr2 == ',');
    ptr = ptr2;
  }

  if(!*arg)
    goto error1;
  else ptr = strchr(ptr, ';');
  if(!ptr) goto error1;

  *ptr = 0;

  free(initcells.cells);

  condcells = cells_from_args(sheet, arg, cell, -1, 0);

  *ptr = ';';

  if(condcells.count <= 0)
    return 0;

  initcells.cells = 0; // to avoid double free if the user presses CTRL+C

  while(1)
  {
    for(n=condcells.count;n--;)
      if(is_true(condcells.cells[n])) goto istrue;
    break;
    istrue:

    if(!(loop_status & FOR_BODYCELLS_KNOWN))
    {
      arg = ptr + 1;
      if(*arg)
      {
        if(*arg != ')')
        {
          bodycells = cells_from_args(sheet, arg, cell, -1, 0);
          if(bodycells.count == -1)
          {
            goto error1;
          }
        }
        else memset(&bodycells, 0, sizeof(struct celllist));
        loop_status |= FOR_BODYCELLS_KNOWN;
        program_status.loop_body_eval = 1;
      }
    }
    else
    {
      for(n=0;n<bodycells.count;n++)
        if(bodycells.cells[n]->type != CELLTYPE_FNC)
        {
          cellerror:
          rc = 0;
          goto end;
        }
        else
        {
          free_for_cells(sheet, &bodycells, &condcells);
          if(
#ifdef DOS
kbhit() ||
#endif
sigint_received || !recalculate_cell(sheet, bodycells.cells[n]))
            goto cellerror;
        }
    }
    for(n=0;n<condcells.count;n++)
    {
      if(!(
#ifdef DOS
kbhit() ||
#endif
sigint_received))
      {
        rc = recalculate_cell(sheet, condcells.cells[n]);
        if(condcells.cells[n]->type == CELLTYPE_FNC && !rc)
          goto cellerror;
      }
      else goto cellerror;
    }
  }
  rc = 1;

  end:
  free(condcells.cells);
  if(loop_status & FOR_BODYCELLS_KNOWN)
    free(bodycells.cells);
  program_status.loop_body_eval = 0;

  return rc;
}

int parse_mov(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist cells = cells_from_args(sheet, arg, cell, 2, 0);
  unsigned n;
  double fres = 0;
  struct cell *srccell;
  int rc;

  if(cells.count == -1) return 0;

  if(cells.count != 2) rc = 0;
  else if(add_refs(&cells, cell)) return 0;
  else
  {
    memcpy(&cells.cells[0]->result, &cells.cells[1]->result, sizeof(union cell_result));
    memcpy(&cell->result, &cells.cells[0]->result, sizeof(union cell_result));
    cell->restype = cells.cells[0]->restype = cells.cells[1]->restype;
    cell->len = cells.cells[0]->len = cells.cells[1]->len;
    rc = 1;
  }
  free(cells.cells);

  return rc;
}

/*int parse_average(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[9];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  double fres = 0;
  struct cell *srccell;
  int rc;

  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->restype == CELLTYPE_INT)
      if(srccell->len == CELLLEN_SHORT)
        fres += srccell->result.shortval;
      else if(srccell->len == CELLLEN_LONG)
        fres += srccell->result.longval;
      else
        fres += srccell->result.llongval;
    else
        fres += srccell->result.doubleval;
  }

  fres /= cells.count;

  free(cells.cells);

  *(double *)resptr = fres;
  cell->restype = CELLTYPE_FLT;

  return 1;
}*/

struct celllist list_celllist;
struct celllist list_sortlist;

void reset_lists(void)
{
  free(list_celllist.cells);
  free(list_sortlist.cells);
  list_celllist.cells = list_sortlist.cells = 0;
  program_status.list_cell_id = program_status.slist_cell_id = -1;
}

void loadcmpval(struct logical_result *cval, struct cell *cell)
{
  cval->value.cellptr = cell;
  cval->flt = (cell->type == CELLTYPE_FLT);
}

int sort_list(struct celllist *list)
{
  struct cell *tempptr;
  struct cell *cmpcell;
  unsigned n;
  unsigned c;
  struct logical_result cval[2];
  struct cmpstrct cmpstrct[2];
  int res;

  memset(&cval, 0, 2*sizeof(struct logical_result));

  for(c=list->count;c-->0;)
  {
    // find the cell of the largest value from group C1:Cc and move it to Cc
    loadcmpval(&cval[1], list->cells[c]);
    for(n=c;n--;)
    {
      loadcmpval(&cval[0], list->cells[n]);
      res = compare(cval, cmpstrct);

      if(res < 0)
      {
        tempptr = list->cells[n];
        list->cells[n] = list->cells[c];
        list->cells[c] = tempptr;
        loadcmpval(&cval[1], list->cells[c]);
      }
    }
  }
}

struct cell *parse_list(struct sheet *sheet, char *arg, unsigned sort, char **endptr)
{
  struct cell *retval;
  struct celllist *list;
  signed *list_cell_id;

  if(!current_for_cell)
  {
    *endptr = 0;
    return 0;
  }

  *endptr = strchr(arg, ')');
  if(*endptr) ++*endptr;

  if(!sort)
  {
    list = &list_celllist;
    list_cell_id = &program_status.list_cell_id;
  }
  else
  {
    list = &list_sortlist;
    list_cell_id = &program_status.slist_cell_id;
  }

  if(*list_cell_id == -1)
  {
    *list = cells_from_args(sheet, arg, current_for_cell, -1, 0);
    if(sort && list->count > 0) sort_list(list);
    *list_cell_id = 0;
  }

  do
  {
    if(*list_cell_id >= list->count)
    {
      free(list->cells);
      list->cells = 0;
      return 0;
    }
    retval = list->cells[(*list_cell_id)++];
  }
  while(retval->type == CELLTYPE_STR);

  return retval;
}

int parse_max(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  signed long long ires = LLONG_MIN;
  double fres;
  struct cell *srccell;

  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->restype > cell->restype)
    {
      cell->restype = srccell->restype;
      fres = ires;
    }
    if(cell->restype == CELLTYPE_INT)
    {
      if(srccell->len == CELLLEN_SHORT)
      {
        if(ires < srccell->result.shortval)
          ires = srccell->result.shortval;
      }
      else if(srccell->len == CELLLEN_LONG)
      {
        if(ires < srccell->result.longval)
          ires = srccell->result.longval;
      }
      else
      {
        if(ires < srccell->result.llongval)
          ires = srccell->result.llongval;
      }
    }
    else
    {
      if(srccell->restype == CELLTYPE_INT)
      {
        if(srccell->len == CELLLEN_SHORT)
        {
          if(fres < srccell->result.shortval)
            fres = srccell->result.shortval;
        }
        else
        {
          if(fres < srccell->result.longval)
            fres = srccell->result.longval;
        }
      }
      else
      {
        if(fres < srccell->result.doubleval)
          fres = srccell->result.doubleval;
      }
    }
  }
  free(cells.cells);

  if(cell->restype == CELLTYPE_INT)
  {
    *(signed long long*)resptr = ires;
    if(ires != (long int)ires)
      cell->len = CELLLEN_LLONG;
    else if(ires != (short)ires)
      cell->len = CELLLEN_LONG;
  }
  else
  {
    *(double *)resptr = fres;
  }

  return 1;
}

int parse_count(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[7];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  signed long long ires = 0;
  struct cell *srccell;

  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->type != CELLTYPE_STR)
      ires++;
  }
  free(cells.cells);

  cell->restype = CELLTYPE_INT;
  cell->len = CELLLEN_SHORT;
  *(signed short*)resptr = ires;

  return 1;  
}

int parse_min(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  signed long long ires = LLONG_MAX;
  double fres;
  struct cell *srccell;

  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->restype > cell->restype)
    {
      cell->restype = srccell->restype;
      fres = ires;
    }
    if(cell->restype == CELLTYPE_INT)
    {
      if(srccell->len == CELLLEN_SHORT)
      {
        if(ires > srccell->result.shortval)
          ires = srccell->result.shortval;
      }
      else if(srccell->len == CELLLEN_LONG)
      {
        if(ires > srccell->result.longval)
          ires = srccell->result.longval;
      }
      else
      {
        if(ires > srccell->result.llongval)
          ires = srccell->result.llongval;
      }
    }
    else
    {
      if(srccell->restype == CELLTYPE_INT)
      {
        if(srccell->len == CELLLEN_SHORT)
        {
          if(fres > srccell->result.shortval)
            fres = srccell->result.shortval;
        }
        else
        {
          if(fres > srccell->result.longval)
            fres = srccell->result.longval;
        }
      }
      else
      {
        if(fres > srccell->result.doubleval)
          fres = srccell->result.doubleval;
      }
    }
  }
  free(cells.cells);

  if(cell->restype == CELLTYPE_INT)
  {
    *(signed long long*)resptr = ires;
    if(ires != (long int)ires)
      cell->len = CELLLEN_LLONG;
    else if(ires != (short)ires)
      cell->len = CELLLEN_LONG;
  }
  else
  {
    *(double *)resptr = fres;
  }

  return 1;
}

int parse_and(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  signed long long ires = -1;
  struct cell *srccell;

  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  cell->restype = CELLTYPE_INT;
  cell->len = CELLLEN_LLONG;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->len < cell->len)
      cell->len = srccell->len;

    if(srccell->len == CELLLEN_SHORT)
      ires &= srccell->result.shortval;
    else if(srccell->len == CELLLEN_LONG)
      ires &= srccell->result.longval;
    else
      ires &= srccell->result.llongval;
  }
  free(cells.cells);

  *(signed long long*)resptr = ires;

  return 1;
}

int parse_or(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[4];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  signed long long ires = 0;
  struct cell *srccell;

  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  cell->restype = CELLTYPE_INT;
  cell->len = CELLLEN_SHORT;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->len > cell->len)
      cell->len = srccell->len;

    if(srccell->len == CELLLEN_SHORT)
      ires |= srccell->result.shortval;
    else if(srccell->len == CELLLEN_LONG)
      ires |= srccell->result.longval;
    else
      ires |= srccell->result.llongval;
  }
  free(cells.cells);

  *(signed long long*)resptr = ires;

  return 1;
}

int parse_xor(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  signed long long ires = 0;
  struct cell *srccell;

  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  cell->restype = CELLTYPE_INT;
  cell->len = CELLLEN_SHORT;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->len > cell->len)
      cell->len = srccell->len;

    if(srccell->len == CELLLEN_SHORT)
      ires ^= srccell->result.shortval;
    else if(srccell->len == CELLLEN_LONG)
      ires ^= srccell->result.longval;
    else
      ires ^= srccell->result.llongval;
  }
  free(cells.cells);

  *(signed long long*)resptr = ires;

  return 1;
}

int parse_not(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist cells = cells_from_args(sheet, arg, cell, 1, 0);
  unsigned n;
  signed long long ires = 0;
  struct cell *srccell;

  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  srccell = cells.cells[0];
  cell->len = srccell->len;

  if(srccell->len == CELLLEN_SHORT)
    ires = ~srccell->result.shortval;
  else if(srccell->len == CELLLEN_LONG)
    ires = ~srccell->result.longval;
  else
    ires = ~srccell->result.llongval;
  free(cells.cells);

  *(signed long long*)resptr = ires;
  cell->restype = CELLTYPE_INT;

  return 1;

}

int parse_product(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[9];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  signed long long ires = 1;
  double fres;
  struct cell *srccell;


  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->restype > cell->restype)
    {
      cell->restype = srccell->restype;
      fres = ires;
    }
    if(cell->restype == CELLTYPE_INT)
    {
      if(srccell->len == CELLLEN_SHORT)
        ires *= srccell->result.shortval;
      else if(srccell->len == CELLLEN_LONG)
        ires *= srccell->result.longval;
      else
        ires *= srccell->result.llongval;
    }
    else
    {
      if(srccell->restype == CELLTYPE_INT)
        if(srccell->len == CELLLEN_SHORT)
          fres *= srccell->result.shortval;
        else if(srccell->len == CELLLEN_LONG)
          fres *= srccell->result.longval;
        else
          fres *= srccell->result.llongval;
      else
          fres *= srccell->result.doubleval;
    }
  }
  free(cells.cells);

  if(cell->restype == CELLTYPE_INT)
  {
    *(signed long long*)resptr = ires;
    if(ires != (long int)ires)
      cell->len = CELLLEN_LLONG;
    else if(ires != (short)ires)
      cell->len = CELLLEN_LONG;
  }
  else
  {
    *(double *)resptr = fres;
  }

  return 1;
}

int parse_imdiv(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[7];
  struct celllist cells = cells_from_args(sheet, arg, cell, 2, 0);
  unsigned n;
  double fres;
  double div;
  struct cell *srccell;


  if(cells.count != 2) return 0;

  if(add_refs(&cells, cell)) return 0;

  srccell = cells.cells[0];

  if(srccell->restype == CELLTYPE_FLT)
    fres = srccell->result.doubleval;
  else if(srccell->len == CELLLEN_SHORT)
    fres = srccell->result.shortval;
  else if(srccell->len == CELLLEN_LONG)
    fres = srccell->result.longval;
  else
    fres = srccell->result.llongval;

  srccell = cells.cells[1];

  if(srccell->restype == CELLTYPE_FLT)
    div = srccell->result.doubleval;
  else if(srccell->len == CELLLEN_SHORT)
    div = srccell->result.shortval;
  else if(srccell->len == CELLLEN_LONG)
    div = srccell->result.longval;
  else
    div = srccell->result.llongval;

  if(!div) return 0;

  fres /= div;

  free(cells.cells);

  *(double*)resptr = fres;
  cell->restype = CELLTYPE_FLT;

  return 1;
}

int parse_quotient(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[10];
  struct celllist cells = cells_from_args(sheet, arg, cell, 2, 0);
  unsigned n;
  signed long long ires;
  signed long long div;
  struct cell *srccell;


  if(cells.count != 2) return 0;

  if(add_refs(&cells, cell)) return 0;

  srccell = cells.cells[0];

  if(srccell->restype == CELLTYPE_FLT)
    ires = srccell->result.doubleval;
  else if(srccell->len == CELLLEN_SHORT)
    ires = srccell->result.shortval;
  else if(srccell->len == CELLLEN_LONG)
    ires = srccell->result.longval;
  else
    ires = srccell->result.llongval;

  srccell = cells.cells[1];

  if(srccell->restype == CELLTYPE_FLT)
    div = srccell->result.doubleval;
  else if(srccell->len == CELLLEN_SHORT)
    div = srccell->result.shortval;
  else if(srccell->len == CELLLEN_LONG)
    div = srccell->result.longval;
  else
    div = srccell->result.llongval;

  if(!div) return 0;

  ires /= div;

  free(cells.cells);

  *(signed long long*)resptr = ires;
  if(ires != (long int)ires)
    cell->len = CELLLEN_LLONG;
  else if(ires != (short)ires)
    cell->len = CELLLEN_LONG;
  cell->restype = CELLTYPE_INT;

  return 1;
}

int parse_mod(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist cells = cells_from_args(sheet, arg, cell, 2, 0);
  unsigned n;
  signed long long ires;
  signed long long div;
  struct cell *srccell;


  if(cells.count != 2) return 0;

  if(add_refs(&cells, cell)) return 0;

  srccell = cells.cells[0];

  if(srccell->restype == CELLTYPE_FLT)
    ires = srccell->result.doubleval;
  else if(srccell->len == CELLLEN_SHORT)
    ires = srccell->result.shortval;
  else if(srccell->len == CELLLEN_LONG)
    ires = srccell->result.longval;
  else
    ires = srccell->result.llongval;

  srccell = cells.cells[1];

  if(srccell->restype == CELLTYPE_FLT)
    div = srccell->result.doubleval;
  else if(srccell->len == CELLLEN_SHORT)
    div = srccell->result.shortval;
  else if(srccell->len == CELLLEN_LONG)
    div = srccell->result.longval;
  else
    div = srccell->result.llongval;

  if(!div) return 0;

  ires %= div;

  free(cells.cells);

  *(signed long long*)resptr = ires;
  if(ires != (long int)ires)
    cell->len = CELLLEN_LLONG;
  else if(ires != (short)ires)
    cell->len = CELLLEN_LONG;
  cell->restype = CELLTYPE_INT;

  return 1;
}

int parse_format(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[8];
  struct celllist cells = cells_from_args(sheet, arg, cell, 1, &arg);
  unsigned n,r,c;
  int rc;
  struct coords coords[2];
  char *ptr = arg;
  struct cell *newcell;
  char a;

  if(cells.count == -1) return 0;

  if(cells.count < 1) rc = 0;
  else
  {
    while(1)
    {
      memset(coords, 0, 2*sizeof(struct coords));
      str_to_coords(arg, &coords[0], &arg);
      switch(a = *arg++)
      {
        case ':':
          str_to_coords(arg, &coords[1], &arg);
          if(*arg) arg++;
          break;
        case ';': case ',': case ')':
          break;
        default:
          goto exitloop;
      }
      if(coords[0].row && coords[0].col)
        if(coords[1].row && coords[1].col) // range of cells
        {
          for(r=coords[0].row;r<=coords[1].row;r++)
            for(c=coords[0].col;c<=coords[1].col;c++)
              if(!(newcell = ptr_to_cell(sheet, c, r)))
              {
                if(add_cell(sheet, c, r, cells.cells[0]->value))
                  goto outofmemory;
              }
              else
              {
                oldcell:
                strcpy(newcell->value, cells.cells[0]->value);
              }
        }
        else // single cell
        {
          if(!(newcell = ptr_to_cell(sheet, coords[0].col, coords[0].row)))
          {
            if(add_cell(sheet, coords[0].col, coords[0].row, cells.cells[0]->value))
            {
              outofmemory:
              rc = 0;
              goto end;
            }
          }
          else goto oldcell;//strcpy(newcell->value, cells.cells[0]->value);
        }
      else if(a != ')')
      {
        rc = 0; // cannot allocate infinite amount of cells...
        goto end;
      }
      else break;
    }
    exitloop:
    cell->result = cells.cells[0]->result;
    cell->restype = cells.cells[0]->restype;
    cell->type = cells.cells[0]->type;
    free(cells.cells);
    cells = cells_from_args(sheet, ptr, cell, -1, 0);
    if(cells.count == -1)
      return 0;
    if(cells.count == 0)
      rc = 0;
    else
    {
//      cell->result = cells.cells[0]->result;
//      cell->restype = cells.cells[0]->restype;
      for(n=0;n<cells.count;n++)
      {
        cells.cells[n]->result = cell->result;
//          cells.cells[n]->type = cells.cells[n]->restype = cells.cells[0]->restype;
        cells.cells[n]->type = cell->type;
        cells.cells[n]->restype = cell->restype;
        cells.cells[n]->len = cell->len;
      }
      rc = 1;
    }
    end:
    free(cells.cells);
  }

  return rc;
}

int parse_sum(struct sheet *sheet, struct cell *cell, void *resptr)
{
  char *arg = &cell->value[5];
  struct celllist cells = cells_from_args(sheet, arg, cell, -1, 0);
  unsigned n;
  signed long long ires = 0;
  double fres;
  struct cell *srccell;


  if(cells.count == -1) return 0;

  if(add_refs(&cells, cell)) return 0;

  for(n=cells.count;n--;)
  {
    srccell = cells.cells[n];
    if(srccell->restype > cell->restype)
    {
      cell->restype = srccell->restype;
      fres = ires;
    }
    if(cell->restype == CELLTYPE_INT)
    {
      if(srccell->len == CELLLEN_SHORT)
        ires += srccell->result.shortval;
      else if(srccell->len == CELLLEN_LONG)
        ires += srccell->result.longval;
      else
        ires += srccell->result.llongval;
    }
    else
    {
      if(srccell->restype == CELLTYPE_INT)
        if(srccell->len == CELLLEN_SHORT)
          fres += srccell->result.shortval;
        else if(srccell->len == CELLLEN_LONG)
          fres += srccell->result.longval;
        else
          fres += srccell->result.llongval;
      else
          fres += srccell->result.doubleval;
    }
  }
  free(cells.cells);

  if(cell->restype == CELLTYPE_INT)
  {
    *(signed long long*)resptr = ires;
    if(ires != (long int)ires)
      cell->len = CELLLEN_LLONG;
    else if(ires != (short)ires)
      cell->len = CELLLEN_LONG;
  }
  else
  {
    *(double *)resptr = fres;
  }

  return 1;
}

int parse_function(struct sheet *sheet, struct cell *cell)
{
  int rc;
  struct cell *old_current_for_cell;
  struct macro *macro;


  if(cell->value[0] != '=')
    return 0;
  else
  {
    if(!strncmp(&cell->value[1], "NOP(", 4))
      rc = 2;
    else
    {
      cell->restype = CELLTYPE_INT;
      cell->len = CELLLEN_SHORT;
      if(!strncmp(&cell->value[1], "SUM(", 4))
        rc = parse_sum(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "MIN(", 4))
      rc = parse_min(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "MAX(", 4))
        rc = parse_max(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "COUNT(", 6))
        rc = parse_count(sheet, cell, &cell->result);
/*    else if(!strncmp(&cell->value[1], "AVERAGE(", 8))
      rc = parse_average(sheet, cell, &cell->result);*/
      else if(!strncmp(&cell->value[1], "IF(", 3))
        rc = parse_if(sheet, cell);
      else if(!strncmp(&cell->value[1], "AND(", 4))
        rc = parse_and(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "OR(", 3))
        rc = parse_or(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "XOR(", 4))
        rc = parse_xor(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "NOT(", 4))
        rc = parse_not(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "PRODUCT(", 8))
        rc = parse_product(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "QUOTIENT(", 8))
        rc = parse_quotient(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "MOD(", 4))
        rc = parse_mod(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "IMDIV(", 6))
        rc = parse_imdiv(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "MOV(", 4))
        rc = parse_mov(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "FORMAT(", 6))
        rc = parse_format(sheet, cell, &cell->result);
      else if(!strncmp(&cell->value[1], "FOR(", 4))
      {
        if(current_for_cell == cell)
        {
          rc = 1;
          goto end;
        }
        else
        {
          old_current_for_cell = current_for_cell;
          current_for_cell = cell;
          rc = parse_for(sheet, cell, &cell->result);
          current_for_cell = old_current_for_cell;
        }
      }
      else if(macro = is_macro(&cell->value[1]))
      {
        rc = run_macro(sheet, cell->value, macro, cell);
      }
      else
      {
        sprintf(statuslinebuff, "Invalid function \"%s\"", &cell->value[1]);
        show_status_getch(statuslinebuff);
        rc = 0;
      }
      if(!rc) cell->error = 1;
    }
  }
  if(rc == 1 && cell->coords.row) save_temporary_deps(sheet, cell);
  end:
//  draw_table(sheet);
//  refresh();

  return rc;
}

int recalculate_cell(struct sheet *sheet, struct cell *cell)
{
  free_deps(sheet, cell);
  return parse_function(sheet, cell);
}

int recalculate_range(struct sheet *sheet, unsigned row)
{
  struct line *line = &sheet->lines[row-1];
  unsigned n, l;
  struct cell *cell;

  for(n=line->range_ref_count;n--;)
  {
    cell = ptr_to_cell(sheet, line->range_refs[n].col, line->range_refs[n].row);
    recalculate_cell(sheet, cell);
  }

  for(l=sheet->rows;l--;)
  {
    line = &sheet->lines[l];
    for(n=line->cell_count;n--;)
      if((cell = line->cells[n]) && line->cells[n]->col_dep)
        recalculate_cell(sheet, cell);
  }
}

void show_calculating()
{
  show_status("Calculating...");
}

int calculate_cell_result(struct sheet *sheet, struct cell *cell)
{
  signed long long intval;
  long double fltval;
  unsigned y = cell->coords.row;
  int rc;

  cell->error = 0;

  if(rc = parse_function(sheet, cell))
    cell->type = CELLTYPE_FNC;
  else if(cell->value[0] >= '0' && cell->value[0] <= '9' || cell->value[0] == '-')
  {
    if(strchr(cell->value, '.'))
    {
      cell->restype = cell->type = CELLTYPE_FLT;
      cell->result.doubleval = atof(cell->value);
    }
    else
    {
      cell->restype = cell->type = CELLTYPE_INT;
      cell->result.longval = atol(cell->value);
      if(cell->result.longval != cell->result.shortval)
        cell->len = CELLLEN_LONG;
    }
  }
  else
    cell->restype = cell->type = CELLTYPE_STR;

  if(y && rc != 2 && !cell->error)
    recalculate_range(sheet, y);

  return rc;
}

int recalculate_refs(struct sheet *sheet, struct cell *cell)
{
  unsigned n;
  struct cell *ref;
  int rc = 0;

#ifdef POSIX
  if(!++parser_callstack)
#else
  if(++parser_callstack == 16)
#endif
  {
    sigint_received = 1;
    cell->error = 1;
  }

  for(n=cell->ref_count;n--;)
  {
    if(
#ifdef DOS
kbhit() ||
#endif
sigint_received)
    {
      show_status_getch("Interrupted.");
      rc = 1;
    }
    else
    {
      ref = ptr_to_cell(sheet, cell->refs[n].col, cell->refs[n].row);
      if(!parse_function(sheet, ref))
      {
        error:
        if(parser_callstack == 1)
        {
          show_status_getch("Reference broken");
        }
        rc = 1;
        break;
      }
      if(recalculate_refs(sheet, ref))
        goto error;
    }
  }

  --parser_callstack;
  return rc;
}
