#define CELLVAL_LEN 80
#define MACRO_LEN 256
#define MACRONAME_LEN 10
#define MAX_ARGS CELLVAL_LEN/3
#define FILENAME_LEN 12

#define CELLTYPE_STR 0
#define CELLTYPE_FNC 1
#define CELLTYPE_INT 2
#define CELLTYPE_FLT 3

#define CELLLEN_SHORT 0
#define CELLLEN_LONG 1
#define CELLLEN_LLONG 2
#define CELLLEN_TOOLONG 3

struct macro
{
  unsigned char max_args;
  unsigned char len;
  char name[MACRONAME_LEN];
  char function[MACRO_LEN];
};

struct var
{
  struct cell *cell;
  char name;
};

struct cmpstrct
{
  signed long long cmpint;
  double cmpflt;
};

union cell_result
{
  signed shortval;
  signed long longval;
  signed long long llongval;
  float floatval;
  double doubleval;
  long double ldoubleval;
};

union logical_eval
{
  struct cell *cellptr;
  long longval;
  double floatval;
};

struct logical_result
{
  union logical_eval value;
  unsigned imm:1;
  unsigned flt:1;
};

struct coords
{
  unsigned row;
  unsigned col;
};

struct cell
{
  union cell_result result;
  struct coords *refs;
  unsigned ref_count;
  struct coords *deps;
  unsigned dep_count;
  struct coords coords;
  unsigned type:2;
  unsigned restype:2;
  unsigned len:2;
  unsigned col_dep:1;
  unsigned error:1;
  unsigned loopvar:1;
  char value[CELLVAL_LEN];
};

struct temporary_cell
{
  union cell_result result;
  struct coords *refs;
  unsigned ref_count;
  struct coords *deps;
  unsigned dep_count;
  struct coords coords;
  unsigned type:2;
  unsigned restype:2;
  unsigned len:2;
  unsigned col_dep:1;
  unsigned error:1;
  unsigned loopvar:1;
  char value[MACRO_LEN];
};


struct celllist
{
  struct cell **cells;
  signed count;
};

struct line
{
  struct coords *range_refs;
  unsigned range_ref_count;
  unsigned cell_count;
  struct cell **cells;
};

struct sheet
{
  unsigned rows;
  struct line *lines;
  unsigned selected_row;
  unsigned selected_col;
  unsigned scroll_row;
  unsigned scroll_col;
  char filename[FILENAME_LEN+1];
};

struct temp_cells
{
  unsigned count;
  struct cell **cells;
};
